@@ -17,7 +17,7 @@ import copy
 from collections import OrderedDict
 
 import pynetbox.core.app
-from six.moves.urllib.parse import urlsplit
+from urllib.parse import urlsplit
 from pynetbox.core.query import Request
 from pynetbox.core.util import Hashabledict
 
@@ -58,53 +58,165 @@ def get_return(lookup, return_fields=None):
 
 
 def flatten_custom(custom_dict):
-    return {
-        k: v if not isinstance(v, dict) else v["value"] for k, v in custom_dict.items()
-    }
+    ret = {}
+
+    for k, val in custom_dict.items():
+        current_val = val
+
+        if isinstance(val, dict):
+            current_val = val.get("id", val)
+
+        if isinstance(val, list):
+            current_val = [v.get("id") if isinstance(v, dict) else v for v in val]
+
+        ret[k] = current_val
+    return ret
 
 
-class JsonField(object):
+class JsonField:
     """Explicit field type for values that are not to be converted
     to a Record object"""
 
     _json_field = True
 
 
-class Record(object):
-    """Create python objects from netbox API responses.
+class RecordSet:
+    """Iterator containing Record objects.
 
-    Creates an object from a NetBox response passed as `values`.
+    Returned by :py:meth:`.Endpoint.all()` and :py:meth:`.Endpoint.filter()` methods.
+    Allows iteration of and actions to be taken on the results from the aforementioned
+    methods. Contains :py:class:`.Record` objects.
+
+    :Examples:
+
+    To see how many results are in a query by calling ``len()``:
+
+    >>> x = nb.dcim.devices.all()
+    >>> len(x)
+    123
+    >>>
+
+    Simple iteration of the results:
+
+    >>> devices = nb.dcim.devices.all()
+    >>> for device in devices:
+    ...     print(device.name)
+    ...
+    test1-leaf1
+    test1-leaf2
+    test1-leaf3
+    >>>
+
+    """
+
+    def __init__(self, endpoint, request, **kwargs):
+        self.endpoint = endpoint
+        self.request = request
+        self.response = self.request.get()
+        self._response_cache = []
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        if self._response_cache:
+            return self.endpoint.return_obj(
+                self._response_cache.pop(), self.endpoint.api, self.endpoint
+            )
+        return self.endpoint.return_obj(
+            next(self.response), self.endpoint.api, self.endpoint
+        )
+
+    def __len__(self):
+        try:
+            return self.request.count
+        except AttributeError:
+            try:
+                self._response_cache.append(next(self.response))
+            except StopIteration:
+                return 0
+            return self.request.count
+
+    def update(self, **kwargs):
+        """Updates kwargs onto all Records in the RecordSet and saves these.
+
+        Updates are only sent to the API if a value were changed, and only for
+        the Records which were changed
+
+        :returns: True if the update succeeded, None if no update were required
+        :example:
+
+        >>> result = nb.dcim.devices.filter(site_id=1).update(status='active')
+        True
+        >>>
+        """
+        updates = []
+        for record in self:
+            # Update each record and determine if anything was updated
+            for k, v in kwargs.items():
+                setattr(record, k, v)
+            record_updates = record.updates()
+            if record_updates:
+                # if updated, add the id to the dict and append to list of updates
+                record_updates["id"] = record.id
+                updates.append(record_updates)
+        if updates:
+            return self.endpoint.update(updates)
+        else:
+            return None
+
+    def delete(self):
+        r"""Bulk deletes objects in a RecordSet.
+
+        Allows for batch deletion of multiple objects in a RecordSet
+
+        :returns: True if bulk DELETE operation was successful.
+
+        :Examples:
+
+        Deleting offline `devices` on site 1:
+
+        >>> netbox.dcim.devices.filter(site_id=1, status="offline").delete()
+        >>>
+        """
+        return self.endpoint.delete(self)
+
+
+class Record:
+    """Create Python objects from NetBox API responses.
+
+    Creates an object from a NetBox response passed as ``values``.
     Nested dicts that represent other endpoints are also turned
-    into Record objects. All fields are then assigned to the
+    into ``Record`` objects. All fields are then assigned to the
     object's attributes. If a missing attr is requested
     (e.g. requesting a field that's only present on a full response on
-    a Record made from a nested response) the pynetbox will make a
-    request for the full object and return the requsted value.
+    a ``Record`` made from a nested response) then pynetbox will make a
+    request for the full object and return the requested value.
 
     :examples:
 
-    Default representation of the object is usually its name
+    Default representation of the object is usually its name:
 
     >>> x = nb.dcim.devices.get(1)
     >>> x
     test1-switch1
     >>>
 
-    Querying a string field.
+    Querying a string field:
 
     >>> x = nb.dcim.devices.get(1)
     >>> x.serial
     'ABC123'
     >>>
 
-    Querying a field on a nested object.
+    Querying a field on a nested object:
 
     >>> x = nb.dcim.devices.get(1)
     >>> x.device_type.model
     'QFX5100-24Q'
     >>>
 
-    Casting the object as a dictionary.
+    Casting the object as a dictionary:
 
     >>> from pprint import pprint
     >>> pprint(dict(x))
@@ -114,7 +226,7 @@ class Record(object):
      'config_context': {},
      'created': '2018-04-01',
      'custom_fields': {},
-     'device_role': {'id': 1,
+     'role': {'id': 1,
                      'name': 'Test Switch',
                      'slug': 'test-switch',
                      'url': 'http://localhost:8000/api/dcim/device-roles/1/'},
@@ -149,7 +261,7 @@ class Record(object):
      'virtual_chassis': None}
      >>>
 
-     Iterating over a Record object.
+     Iterating over a ``Record`` object:
 
     >>> for i in x:
     ...  print(i)
@@ -211,7 +323,12 @@ class Record(object):
         return dict(self)[k]
 
     def __str__(self):
-        return getattr(self, "name", None) or getattr(self, "label", None) or ""
+        return (
+            getattr(self, "name", None)
+            or getattr(self, "label", None)
+            or getattr(self, "display", None)
+            or ""
+        )
 
     def __repr__(self):
         return str(self)
@@ -238,21 +355,42 @@ class Record(object):
 
     def _add_cache(self, item):
         key, value = item
-        if key == "local_context_data":
-            self._init_cache.append((key, copy.deepcopy(value)))
-        else:
-            self._init_cache.append((key, get_return(value)))
+        self._init_cache.append((key, get_return(value)))
 
     def _parse_values(self, values):
-        """ Parses values init arg.
+        """Parses values init arg.
 
         Parses values dict at init and sets object attributes with the
         values within.
         """
 
-        def list_parser(list_item):
+        def generic_list_parser(key_name, list_item):
+            from pynetbox.models.mapper import CONTENT_TYPE_MAPPER
+
+            if (
+                isinstance(list_item, dict)
+                and "object_type" in list_item
+                and "object" in list_item
+            ):
+                lookup = list_item["object_type"]
+                model = None
+                model = CONTENT_TYPE_MAPPER.get(lookup)
+                if model:
+                    return model(list_item["object"], self.api, self.endpoint)
+
+            return list_item
+
+        def list_parser(key_name, list_item):
             if isinstance(list_item, dict):
-                return self.default_ret(list_item, self.api, self.endpoint)
+                lookup = getattr(self.__class__, key_name, None)
+                if not isinstance(lookup, list):
+                    # This is *list_parser*, so if the custom model field is not
+                    # a list (or is not defined), just return the default model
+                    return self.default_ret(list_item, self.api, self.endpoint)
+                else:
+                    model = lookup[0]
+                    return model(list_item, self.api, self.endpoint)
+
             return list_item
 
         for k, v in values.items():
@@ -261,7 +399,7 @@ class Record(object):
                 if k in ["custom_fields", "local_context_data"] or hasattr(
                     lookup, "_json_field"
                 ):
-                    self._add_cache((k, v.copy()))
+                    self._add_cache((k, copy.deepcopy(v)))
                     setattr(self, k, v)
                     continue
                 if lookup:
@@ -271,8 +409,16 @@ class Record(object):
                 self._add_cache((k, v))
 
             elif isinstance(v, list):
-                v = [list_parser(i) for i in v]
-                to_cache = list(v)
+                # check if GFK
+                if len(v) and isinstance(v[0], dict) and "object_type" in v[0]:
+                    v = [generic_list_parser(k, i) for i in v]
+                    to_cache = list(v)
+                elif k == "constraints":
+                    # Permissions constraints can be either dict or list
+                    to_cache = copy.deepcopy(v)
+                else:
+                    v = [list_parser(k, i) for i in v]
+                    to_cache = list(v)
                 self._add_cache((k, to_cache))
 
             else:
@@ -288,8 +434,8 @@ class Record(object):
             url_path = url_path[len(extra_path) :]
         split_url_path = url_path.split("/")
         if split_url_path[2] == "plugins":
-            # Skip plugins from the path
-            app, name = split_url_path[3:5]
+            app = "plugins/{}".format(split_url_path[3])
+            name = split_url_path[4]
         else:
             app, name = split_url_path[2:4]
         return getattr(pynetbox.core.app.App(self.api, app), name)
@@ -308,10 +454,9 @@ class Record(object):
             req = Request(
                 base=self.url,
                 token=self.api.token,
-                session_key=self.api.session_key,
                 http_session=self.api.http_session,
             )
-            self._parse_values(req.get())
+            self._parse_values(next(req.get()))
             self.has_details = True
             return True
         return False
@@ -374,6 +519,30 @@ class Record(object):
         )
         return set([i[0] for i in set(current.items()) ^ set(init.items())])
 
+    def updates(self):
+        """Compiles changes for an existing object into a dict.
+
+        Takes a diff between the objects current state and its state at init
+        and returns them as a dictionary, which will be empty if no changes.
+
+        :returns: dict.
+        :example:
+
+        >>> x = nb.dcim.devices.get(name='test1-a3-tor1b')
+        >>> x.serial
+        ''
+        >>> x.serial = '1234'
+        >>> x.updates()
+        {'serial': '1234'}
+        >>>
+        """
+        if self.id:
+            diff = self._diff()
+            if diff:
+                serialized = self.serialize()
+                return {i: serialized[i] for i in diff}
+        return {}
+
     def save(self):
         """Saves changes to an existing object.
 
@@ -385,26 +554,22 @@ class Record(object):
 
         >>> x = nb.dcim.devices.get(name='test1-a3-tor1b')
         >>> x.serial
-        u''
+        ''
         >>> x.serial = '1234'
         >>> x.save()
         True
         >>>
         """
-        if self.id:
-            diff = self._diff()
-            if diff:
-                serialized = self.serialize()
-                req = Request(
-                    key=self.id,
-                    base=self.endpoint.url,
-                    token=self.api.token,
-                    session_key=self.api.session_key,
-                    http_session=self.api.http_session,
-                )
-                if req.patch({i: serialized[i] for i in diff}):
-                    return True
-
+        updates = self.updates()
+        if updates:
+            req = Request(
+                key=self.id,
+                base=self.endpoint.url,
+                token=self.api.token,
+                http_session=self.api.http_session,
+            )
+            if req.patch(updates):
+                return True
         return False
 
     def update(self, data):
@@ -447,7 +612,6 @@ class Record(object):
             key=self.id,
             base=self.endpoint.url,
             token=self.api.token,
-            session_key=self.api.session_key,
             http_session=self.api.http_session,
         )
         return True if req.delete() else False
